/*
Copyright 2022 Huawei Cloud Computing Technologies Co., Ltd.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package config

import (
	"crypto/tls"
	"fmt"
	"math"
	"runtime"
	"time"

	"github.com/influxdata/influxdb/pkg/tlsconfig"
	"github.com/influxdata/influxdb/services/retention"
	"github.com/influxdata/influxdb/toml"
	"github.com/openGemini/openGemini/lib/errno"
	"github.com/openGemini/openGemini/lib/memory"
	httpdConf "github.com/openGemini/openGemini/open_src/influx/httpd/config"
	"github.com/openGemini/openGemini/services/stream"
)

const (
	EngineType1                         = "tssp1"
	EngineType2                         = "tssp2"
	DefaultEngine                       = "tssp1"
	DefaultImmutableMaxMemoryPercent    = 10
	DefaultCompactFullWriteColdDuration = 1 * time.Hour

	KB = 1024
	MB = 1024 * 1024
	GB = 1024 * 1024 * 1024

	// DefaultMaxConcurrentCompactions is the maximum number of concurrent full and level compactions
	// that can run at one time.  A value of 0 results in 50% of runtime.GOMAXPROCS(0) used at runtime.
	DefaultMaxConcurrentCompactions = 0

	DefaultWriteColdDuration = 5 * time.Second
	Is32BitPtr               = (^uintptr(0) >> 32) == 0

	DefaultSnapshotThroughput      = 48 * MB
	DefaultSnapshotThroughputBurst = 64 * MB
	DefaultMaxWriteHangTime        = 15 * time.Second
	DefaultWALSyncInterval         = 100 * time.Millisecond
	DefaultReadCachePercent        = 3
	MinFlavorMemory                = 8 * GB
	MaxFlavorMemory                = 512 * GB

	IndexFileDirectory = "index"
	DataDirectory      = "data"
	WalDirectory       = "wal"
)

// TSStore represents the configuration format for the influxd binary.
type TSStore struct {
	Common      *Common     `toml:"common"`
	Data        Store       `toml:"data"`
	Coordinator Coordinator `toml:"coordinator"`
	Monitor     Monitor     `toml:"monitor"`
	Logging     Logger      `toml:"logging"`
	Gossip      *Gossip     `toml:"gossip"`
	Spdy        Spdy        `toml:"spdy"`

	HTTPD             httpdConf.Config `toml:"http"`
	Retention         retention.Config `toml:"retention"`
	DownSample        retention.Config `toml:"downsample"`
	HierarchicalStore retention.Config `toml:"hierarchical-storage"`
	Stream            stream.Config    `toml:"stream"`

	// TLS provides configuration options for all https endpoints.
	TLS      tlsconfig.Config `toml:"tls"`
	Analysis Castor           `toml:"castor"`
	Sherlock *SherlockConfig  `toml:"sherlock"`
}

// NewTSStore returns an instance of Config with reasonable defaults.
func NewTSStore() *TSStore {
	c := &TSStore{}

	c.Common = NewCommon()
	c.Data = NewStore()
	c.Coordinator = NewCoordinator()

	c.Monitor = NewMonitor(AppStore)
	c.HTTPD = httpdConf.NewConfig()
	c.Logging = NewLogger(AppStore)

	c.Retention = retention.NewConfig()
	c.DownSample = retention.NewConfig()
	c.HierarchicalStore = retention.NewConfig()
	c.Gossip = NewGossip()

	c.Analysis = NewCastor()
	c.Stream = stream.NewConfig()
	c.Sherlock = NewSherlockConfig()
	return c
}

// Validate returns an error if the config is invalid.
func (c *TSStore) Validate() error {
	if c.Data.Engine != EngineType1 && c.Data.Engine != EngineType2 {
		return fmt.Errorf("unknown tsm engine type: %v", c.Data.Engine)
	}

	items := []Validator{
		c.Common,
		c.Data,
		c.Monitor,
		c.Retention,
		c.DownSample,
		c.HierarchicalStore,
		c.TLS,
		c.Logging,
		c.Spdy,
		c.Analysis,
		c.Sherlock,
	}

	for _, item := range items {
		if err := item.Validate(); err != nil {
			return err
		}
	}

	return nil
}

// ApplyEnvOverrides apply the environment configuration on top of the config.
func (c *TSStore) ApplyEnvOverrides(fn func(string) string) error {
	return toml.ApplyEnvOverrides(fn, "TSSTORE", c)
}

func (c *TSStore) GetLogging() *Logger {
	return &c.Logging
}

func (c *TSStore) GetSpdy() *Spdy {
	return &c.Spdy
}

func (c *TSStore) GetCommon() *Common {
	return c.Common
}

// Store is the configuration for the engine.
type Store struct {
	IngesterAddress string      `toml:"store-ingest-addr"`
	SelectAddress   string      `toml:"store-select-addr"`
	Domain          string      `toml:"domain"`
	TLS             *tls.Config `toml:"-"`
	DataDir         string      `toml:"store-data-dir"`
	WALDir          string      `toml:"store-wal-dir"`
	MetaDir         string      `toml:"store-meta-dir"`
	Engine          string      `toml:"engine-type"`
	Index           string      `toml:"index-version"`
	OpsMonitor      *OpsMonitor `toml:"ops-monitor"`
	// The max inmem percent of immutable
	ImmTableMaxMemoryPercentage int `toml:"imm-table-max-memory-percentage"`

	// Compaction options for tssp1 (descriptions above with defaults)
	CompactFullWriteColdDuration toml.Duration `toml:"compact-full-write-cold-duration"`
	MaxConcurrentCompactions     int           `toml:"max-concurrent-compactions"`
	MaxFullCompactions           int           `toml:"max-full-compactions"`
	CompactThroughput            toml.Size     `toml:"compact-throughput"`
	CompactThroughputBurst       toml.Size     `toml:"compact-throughput-burst"`
	SnapshotThroughput           toml.Size     `toml:"snapshot-throughput"`
	SnapshotThroughputBurst      toml.Size     `toml:"snapshot-throughput-burst"`
	CompactionMethod             int           `toml:"compaction-method"` // 0:auto, 1: streaming, 2: non-streaming

	// Configs for snapshot
	WriteColdDuration     toml.Duration `toml:"write-cold-duration"`
	ShardMutableSizeLimit toml.Size     `toml:"shard-mutable-size-limit"`
	NodeMutableSizeLimit  toml.Size     `toml:"node-mutable-size-limit"`
	MaxWriteHangTime      toml.Duration `toml:"max-write-hang-time"`
	MemDataReadEnabled    bool          `toml:"mem-data-read-enabled"`

	WalSyncInterval   toml.Duration `toml:"wal-sync-interval"`
	WalEnabled        bool          `toml:"wal-enabled"`
	WalReplayParallel bool          `toml:"wal-replay-parallel"`
	WalReplayAsync    bool          `toml:"wal-replay-async"`
	CacheDataBlock    bool          `toml:"cache-table-data-block"`
	CacheMetaBlock    bool          `toml:"cache-table-meta-block"`
	EnableMmapRead    bool          `toml:"enable-mmap-read"`
	Readonly          bool          `toml:"readonly"`
	CompactRecovery   bool          `toml:"compact-recovery"`
	QuerySeriesLimit  int           `toml:"query-series-limit"`

	ReadCacheLimit       toml.Size `toml:"read-cache-limit"`
	WriteConcurrentLimit int       `toml:"write-concurrent-limit"`
	OpenShardLimit       int       `toml:"open-shard-limit"`

	DownSampleWriteDrop bool `toml:"downsample-write-drop"`
}

// NewStore returns the default configuration for tsdb.
func NewStore() Store {
	size, _ := memory.SysMem()
	memorySize := toml.Size(size * KB)
	readCacheLimit := getCacheLimitSize(uint64(memorySize))
	return Store{
		Engine:                       DefaultEngine,
		ImmTableMaxMemoryPercentage:  DefaultImmutableMaxMemoryPercent,
		CompactFullWriteColdDuration: toml.Duration(DefaultCompactFullWriteColdDuration),
		MaxConcurrentCompactions:     DefaultMaxConcurrentCompactions,
		MaxFullCompactions:           1,
		SnapshotThroughput:           toml.Size(DefaultSnapshotThroughput),
		SnapshotThroughputBurst:      toml.Size(DefaultSnapshotThroughputBurst),
		WriteColdDuration:            toml.Duration(DefaultWriteColdDuration),
		MaxWriteHangTime:             toml.Duration(DefaultMaxWriteHangTime),
		MemDataReadEnabled:           true,
		CacheDataBlock:               false,
		CacheMetaBlock:               false,
		EnableMmapRead:               !Is32BitPtr,
		ReadCacheLimit:               toml.Size(readCacheLimit),
		WriteConcurrentLimit:         0,
		WalSyncInterval:              toml.Duration(DefaultWALSyncInterval),
		WalEnabled:                   true,
		WalReplayParallel:            false,
		WalReplayAsync:               false,
		CompactRecovery:              true,
		QuerySeriesLimit:             0,
		CompactionMethod:             0,
		OpsMonitor:                   NewOpsMonitorConfig(),
		OpenShardLimit:               0,
		DownSampleWriteDrop:          true,
	}
}

func (c *Store) Corrector(cpuNum int, memorySize toml.Size) {
	if cpuNum <= 0 {
		cpuNum = runtime.NumCPU()
	}
	if memorySize == 0 {
		size, _ := memory.SysMem()
		memorySize = toml.Size(size * KB)
	}
	if c.OpenShardLimit <= 0 {
		c.OpenShardLimit = cpuNum
	}

	if c.ReadCacheLimit != 0 {
		c.ReadCacheLimit = toml.Size(getCacheLimitSize(uint64Limit(8*GB, 512*GB, uint64(memorySize))))
	}
	defaultShardMutableSizeLimit := toml.Size(uint64Limit(8*MB, 1*GB, uint64(memorySize/256)))
	defaultNodeMutableSizeLimit := toml.Size(uint64Limit(32*MB, 16*GB, uint64(memorySize/16)))
	defaultCompactThroughput := toml.Size(uint64Limit(32*MB, 256*MB, uint64(cpuNum*12*MB)))
	defaultCompactThroughputBurst := toml.Size(uint64Limit(32*MB, 384*MB, uint64(cpuNum*15*MB)))

	items := [][2]*toml.Size{
		{&c.ShardMutableSizeLimit, &defaultShardMutableSizeLimit},
		{&c.NodeMutableSizeLimit, &defaultNodeMutableSizeLimit},
		{&c.CompactThroughput, &defaultCompactThroughput},
		{&c.CompactThroughputBurst, &defaultCompactThroughputBurst},
	}

	for i := range items {
		if *items[i][0] == 0 {
			*items[i][0] = *items[i][1]
		}
	}
}

// Validate validates the configuration hold by c.
func (c Store) Validate() error {
	svItems := []stringValidatorItem{
		{"data store-ingest-addr", c.IngesterAddress},
		{"data store-select-addr", c.SelectAddress},
		{"data store-data-dir", c.DataDir},
		{"data store-meta-dir", c.MetaDir},
	}
	if c.WalEnabled {
		svItems = append(svItems, stringValidatorItem{"data store-wal-dir", c.WALDir})
	}
	if err := (stringValidator{}).Validate(svItems); err != nil {
		return err
	}

	ivItems := []intValidatorItem{
		{"data max-concurrent-compactions", int64(c.MaxConcurrentCompactions), true},
		{"data max-full-compactions", int64(c.MaxFullCompactions), true},
		{"data imm-table-max-memory-percentage", int64(c.ImmTableMaxMemoryPercentage), false},
		{"data write-cold-duration", int64(c.WriteColdDuration), false},
		{"data max-write-hang-time", int64(c.MaxWriteHangTime), false},
	}
	iv := intValidator{0, math.MaxInt64}
	if err := iv.Validate(ivItems); err != nil {
		return err
	}

	return nil
}

func (c Store) ValidateEngine(engines []string) error {
	for _, e := range engines {
		if e == c.Engine {
			return nil
		}
	}

	return errno.NewError(errno.UnrecognizedEngine, c.Engine)
}

func (c *Store) InsertAddr() string {
	return CombineDomain(c.Domain, c.IngesterAddress)
}

func (c *Store) SelectAddr() string {
	return CombineDomain(c.Domain, c.SelectAddress)
}

func uint64Limit(min, max uint64, v uint64) uint64 {
	if v < min {
		return min
	}
	if v > max {
		return max
	}
	return v
}

type OpsMonitor struct {
	HttpAddress      string `toml:"store-http-addr"`
	AuthEnabled      bool   `toml:"auth-enabled"`
	HttpsEnabled     bool   `toml:"store-https-enabled"`
	HttpsCertificate string `toml:"store-https-certificate"`
}

func NewOpsMonitorConfig() *OpsMonitor {
	return &OpsMonitor{
		HttpAddress: "",
	}
}

func getCacheLimitSize(size uint64) uint64 {
	return size * DefaultReadCachePercent / 100
}
